/*-
 * Copyright (C) @BABOLO  2002 Dec 08
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 08\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: binfout.c,v 1.5 2007/07/16 15:47:13 gd Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "binfout.h"

BLIN_flag v_flag;
u_int64_t bin[FORMSIZE];

u_int64_t masks[] =
{ 0x5555555555555555ULL, 0x3333333333333333ULL, 0x0F0F0F0F0F0F0F0FULL
, 0x00FF00FF00FF00FFULL, 0x0000FFFF0000FFFFULL, 0x00000000FFFFFFFFULL
};
u_char shifts[] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20};

void
usage(int ex) {
    fprintf(stderr, "binfout " VERS "  " DATE "\n"
"Usage: binfout [-v[v]..] FormatFile\n"
"");
    exit(ex);
}

int
main(int argc, char *argv[]) {
    int in = fileno(stdin);
    FILE *out = stdout;
    int c;
    u_char *format, *buf;
    size_t i, j, last;
    off_t pos;
    struct tnode *root, *leaf;
    mife_descriptor *mf, *fm;
    struct formades *inf;

    v_flag = BLIN_VER1;
    while ((c = getopt(argc, argv, "qv?")) != -1) {
        switch (c) {
        case 'q': setBLIN_VERMASK(v_flag, 0);
                  break;
        case 'v': BLIN_VERBOSE(v_flag);
                  break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc < 1 || !argv || !*argv || !**argv) {
        ifBLIN_QV2(v_flag) warnx("No FormatFile given\n");
        exit(EX_USAGE);
    }
    if  (argc > 2) {
        ifBLIN_QV2(v_flag) warnx("Overaggumented\n");
        exit(EX_USAGE);
    }
    if  (!(fm = mife_open(MIFE_FULL, *argv))) {
        ifBLIN_QV2(v_flag) warn("file %s:", *argv);
        exit(EX_NOINPUT);
    }
    if  (mife_read(fm, 0, 0) < 0) {
        ifBLIN_QV2(v_flag) warn("file %s:", *argv);
        exit(EX_DATAERR);
    }
    if  (!(format = mife_get(fm, 0))) {
        ifBLIN_QV2(v_flag) warn("file %s:", *argv);
        exit(EX_DATAERR);
    }
    inf = informat(&format);
    root = outformat(format);
    if  (argc == 2 && !argv[1] && !*argv[1]) {
        ifBLIN_QV2(v_flag) warnx("No InputFile given\n");
        exit(EX_USAGE);
    }
    if  (argc == 2 && (in = open(argv[1], O_RDONLY)) < 0) {
        ifBLIN_QV2(v_flag) warn("file %s:", argv[1]);
        exit(EX_NOINPUT);
    }
    if  (!(mf = mife_opef(0, in))) {
        ifBLIN_QV2(v_flag) warn("infile %s:", *argv);
        exit(EX_NOINPUT);
    }

    for ( pos = 0
        ; (last = mife_read(mf, inf->tot >> 3, pos)) >= inf->tot >> 3
        ; pos += inf->tot >> 3
        ) {
        buf = mife_get(mf, pos);
        ifBLIN_QV4(v_flag) for (i = 0; i < (inf->tot >> 3); i++) fprintf(stderr, "%02X", buf[i]);
        ifBLIN_QV4(v_flag) fprintf(stderr, "\n");
        for (i = 0, j = 0; i < inf->cur; i++) {
            for (bin[i] = 0, c = 0; c < (inf->form[i].ll >> 3); c++) {
                ifBLIN_QV5(v_flag) fprintf(stderr, "i=%2d j=%3d c=%d f=%02X\n", i, j, c, buf[j]);
                bin[i] = (bin[i] << 8) | (buf[j++] & 0xFF);
            }
            ifBLIN_QV5(v_flag) fprintf(stderr, "bin[%d]=%016llX\n", i, bin[i]);
            for (c = 5; c >= 0; c--) {
                if  (shifts[c] & inf->form[i].ff) {
                    ifBLIN_QV6(v_flag)
                        fprintf( stderr, "c=%d %016llX %016llX %016llX"
                               , c, masks[c]
                               , ((bin[i] & masks[c]) << shifts[c])
                               , ((bin[i] >> shifts[c]) & masks[c])
                               );
                    bin[i] = ((bin[i] & masks[c]) << shifts[c]) | ((bin[i] >> shifts[c]) & masks[c]);
                    ifBLIN_QV6(v_flag)
                        fprintf( stderr, " %016llX\nlen=%02X diff=%02X shift=%02X %02X"
                               , bin[i]
                               , inf->form[i].ll
                               , 64 - inf->form[i].ll
                               , shifts[c]
                               , (64 - inf->form[i].ll) & shifts[c]
                               );
                    if  ((64 - inf->form[i].ll) & shifts[c]) bin[i] = bin[i] >> shifts[c];
                    ifBLIN_QV6(v_flag) fprintf(stderr, " %016llX\n", bin[i]);
            }   }
            ifBLIN_QV4(v_flag) fprintf( stderr, "bin[%d]=%016llX\n", i, bin[i]);
        }
        for (leaf = root; leaf; leaf = leaf->next) {
            ifBLIN_QV5(v_flag) {
                fprintf(stderr, "%2d %02X ~%s~", leaf->n, leaf->ff, leaf->f);
                if  (leaf->ff) fprintf(stderr, leaf->f, bin[leaf->n]);
                  else fprintf(stderr, leaf->f);
            }else{
                if  (leaf->ff) fprintf(out, leaf->f, bin[leaf->n]);
                  else fprintf(out, leaf->f);
    }   }   }
    if  (last < 0) {
        ifBLIN_QV2(v_flag) warn("input");
        exit(EX_DATAERR);
    }
    if (last != 0) {
        ifBLIN_QV2(v_flag) warnx("truncated");
        exit(EX_DATAERR);
    }
    mife_close(mf);
    mife_close(fm);
    exit(EX_OK);
}
