/*-
 * Copyright (C) @BABOLO  2002 Dec 08
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 08\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: outformat.c,v 1.3 2004/03/06 03:34:43 babolo Exp $";
#endif /* not lint */

#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "binfout.h"

#define C_er 0x80000000 /*                                         */
#define C_mb 0x40000000 /*                                  */
#define C_mm 0x20000000 /*                                     */
#define C_mc 0x10000000 /*  next                              */
#define C_fl  0x8000000 /*     - %%                */
#define C_fs  0x4000000 /*                                        */
#define C_fc  0x2000000 /*                                     */
#define C_fo  0x1000000 /*                                       */
#define C_dc   0x800000 /*                                       */
#define C_dg   0x400000 /*                                         */
#define C_de   0x200000 /*  -       */
#define C_ce   0x100000 /*                                          */
#define C_fb    0x80000
#define C_state    0xFF /*                                    */
#define C_mbc (C_mb|C_mc)
#define C_mmc (C_mm|C_mc)
#define C_fsc (C_fs|C_fc)
#define C_fso (C_fs|C_fo)

enum classes {nl, pc, td, dg, te, ok, xx};
/* nl \0
   pc %
   td ~
   te format char
   ok
   xx
 */

enum states {c0, c1, n0, n1, f0, f1, m0, m1, ex};
/* c0 - 
   c1 -    
   n0 -   
   n1 -  
   f0 - 
   f1 -   
   m0 -  
   m1 - 
   ex
 */

struct tnode *
outformat(u_char *format) {
    int c = 0;
    enum states state;
    u_int32_t control = 0;
    struct tnode *root, *leaf;

    const enum classes class[256] = {
     nl,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,pc,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,dg,dg,dg,dg ,dg,dg,dg,dg ,dg,dg,ok,ok ,ok,ok,ok,ok

    ,ok,ok,ok,ok ,ok,te,ok,te ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,te,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,te ,te,te,te,te ,ok,te,ok,ok ,ok,ok,ok,te
    ,ok,ok,ok,ok ,ok,te,ok,ok ,te,ok,ok,ok ,ok,ok,td,ok

    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok

    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok ,ok,ok,ok,ok
    };

    const u_int32_t automa[ex][xx] = {
    /*     \0             %               ~                   dg              te              ok */
     {     ex, C_mmc|C_dc|n0, C_fsc|C_mmc|c1,     C_fsc|C_mmc|c1, C_fsc|C_mmc|c1, C_fsc|C_mmc|c1}/*c0*/
    ,{     ex,       C_dc|m0,             c1,                 c1,             c1,             c1}/*c1*/
    ,{C_er|ex,  C_fl|C_fc|c1,        C_er|ex,            C_dg|n1,        C_er|ex,        C_er|ex}/*n0*/
    ,{C_er|ex,       C_er|ex,  C_de|C_fso|f0,            C_dg|n1,        C_er|ex,        C_er|ex}/*n1*/
    ,{C_er|ex,       C_er|ex,        C_er|ex,                 f0,             f1,             f0}/*f0*/
    ,{     ex,       C_dc|m0,             f1,                 f1,             f1,             f1}/*f1*/
    ,{C_er|ex,  C_fl|C_fc|f1,        C_er|ex, C_mbc|C_ce|C_dg|m1,        C_er|ex,        C_er|ex}/*m0*/
    ,{C_er|ex,       C_er|ex,  C_de|C_fso|f0,            C_dg|m1,        C_er|ex,        C_er|ex}/*m1*/
    };

    leaf = root = NULL;
    for (state = c0; state < ex; format++) {
        control = automa[state][class[*format]];
        ifBLIN_QV3(v_flag) fprintf(stderr, "%08X %02X c=%d\n", control, *format, c);
        if  (control & C_er) exit(EX_DATAERR);
        if  (control & C_mb) leaf = leaf->next = malloc(sizeof(struct tnode));
        if  (control & C_mm) root = leaf = malloc(sizeof(struct tnode));
        if  (control & C_mc) leaf->next = NULL;
        if  (control & C_fl) leaf->f = format - 1;
        if  (control & C_fs) leaf->f = format;
        if  (control & C_fc) leaf->ff = 0;
        if  (control & C_fo) leaf->ff = 1;
        if  (control & C_dc) c = 0;
        if  (control & C_dg) c = c * 10 + (*format - '0');
        if  (control & C_de) {leaf->n = c; *format = '%';}
        if  (control & C_ce) *(format - 1) = '\0';
        state = control & C_state;
    }
    ifBLIN_QV2(v_flag) {
        for (leaf = root; leaf; leaf = leaf->next) {
            if  (leaf->ff) fprintf(stderr, "%2d %02X ~%s~\n", leaf->n, leaf->ff, leaf->f);
              else fprintf(stderr, "   %02X ~%s~\n", leaf->ff, leaf->f);
    }   }
    return(root);
}
